/*
 * parse.c
 * command parsing, argument parsing routines for cancan
 * anyone care to rewrite parse(), subst_param() nonrecursive?
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "cancan.h"
#include "list.h"
#include "term.h"
#include "parse.h"
#include "cmd.h"

FILE *popen();
int pclose();
long strtol();

static void subst_param();
#ifdef XEXPR
static char *get_tok();
static char *brace_var();
#endif
static char *getval();
static char *calc();
static struct param *match_action();
static void split_arg();
static char *skip_arg();

#ifdef OWN_CTYPE
const unsigned char ct_array[256] = {
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     1, 1, 1, 1, 1, 1, 1, 1,33,17, 1, 1, 1, 1, 1, 1,
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 1,65, 1, 1, 1, 1,
     1, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,33, 1,17, 1, 9,
     1, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,33, 1,17, 1
};
#endif

/*
 * silent = 1 if the line should not be echoed, 0 otherwise.
 * if invoked by alias or action, params are passed along.
 */
void parse(cmdline, silent, param)
char cmdline[LINELEN];
int silent;
struct param param[10];
{
    static defnode *recursed_aliases;
    char cmd[LINELEN], buf[LINELEN];
    char *s, *d;
    int i, subst;
    aliasnode *np;

    if(!brace_balance(cmdline)) return;

    s = skipspace(cmdline);			/* line enclosed in braces? */
    if(*s == '{' && !*skipspace((d = skip_arg(++s, '}')) + 1))
	memcpy(cmd, s, d - s), cmd[d - s] = 0, subst = 1;
    else
	strcpy(cmd, cmdline), subst = 0;

    /* PROCESS THE COMMANDS ONE BY ONE */
    s = cmd;
    do {
	cmdline = chop_arg(s, CMDSEP);		/* chop into commands */
	if(recording_name && *s != '#' && !param) record_line(s);

	if((d = skip_arg(s, ' '))) *d = 0;	/* pick out the first word */

	if((np = *lookup_alias(s)) && (*lookup(s, &recursed_aliases) ||
		(recording_name && !strcmp(s, recording_name)))) np = 0;

	if(np) { if(d) *d = ' '; }
	else if(*s == '#') { if(d) d = skipspace(d + 1); }
	else { if(d) *d = ' '; d = s; }		/* restore line */

	buf[0] = buf[sizeof(buf) - 1] = 0;
	if(d)
	    if(subst) subst_param(buf, d, param);
	    else strcpy(buf, d);

	if(param && !cmdline && !subst && recursed_aliases && param[0].len) {
	    i = strlen(buf);		/* append arguments */
	    strncpy(buf + i, param[0].s, sizeof(buf) - i);
	}
	if(buf[sizeof(buf) - 1]) {
	    buf[sizeof(buf) - 1] = 0;
	    printf("#warning: line truncated\n");
	}

	if(np) {				/* process an alias */
	    struct param new_param[10];
	    defnode ban_alias;
	    ban_alias.next = recursed_aliases;  ban_alias.field = s;
	    if(d) *d = 0;			/* cut out first word again */
	    split_arg(buf, new_param);
	    recursed_aliases = &ban_alias;	/* block alias recursion */
	    parse(np->subst, silent, new_param);
	    recursed_aliases = ban_alias.next;
	} else if(*s != '#') {			/* text to send		*/
	    if(buf[0] && (!buf[1] || speedwalk) && parse_walk(buf)) continue;
	    if(!subst) unescape(buf);
	    if(!silent) printf("[%s%s%s]\n", edattrbeg, buf, edattrend);
	    buf[i = strlen(buf)] = '\n';
	    send_to_host(buf, i + 1);
	} else if(*++s) {			/* "# " is a comment	*/
	    if(isdigit(*s) && !*skipnum(s)) {
		if((i = atoi(s)) > 0)
		    while(i--) parse(buf, 1, (struct param *)0);
		else 
		    printf("#bogus repeat count\n");
	    } else
#ifndef OMIT_SESSIONS
	    if(*s == '#') {			/* a session is specified */
		int backfd = sockfd;
		int back_data_sent = data_sent;
		if((sockfd = find_sess(s + 1, 1))) {
		    parse(buf, 1, (struct param *)0);
		    if(!is_main_sess())
			data_sent = back_data_sent;
		}
		sockfd = backfd;
	    } else if(!is_main_sess())
		printf("#command not allowed on subsidiary connection\n");
	    else
#endif
#ifndef NOSHELL
	    if(*s == '?' && !*(s + 1)) {	/* #? command	*/
		i = execute(buf);
		if(!(s = cmdline)) break;
		cmdline = chop_arg(s, CMDSEP);
		if(!i) {
		    parse(s, silent, param);
		    return; 
		}
	    } else
#endif
	    if(param && !recursed_aliases && !strcmp(s, "add")) {
		if(param[0].len) {
		    sprintf(buf, "%.*s", param[0].len, param[0].s);
		    put_word(buf);
		}
	    } else
	    for(i = 0; ; i++)
		if(!commands[i].name) {
		    printf("#unknown cancan command '%s'\n", s);
		    break;
		} else if(commands[i].funct && !strcmp(s, commands[i].name)) {
		    (*commands[i].funct)(buf);
		    if(line0 < lines_1) lastline();
		    break;
		}
	}
    } while((s = cmdline));
}

/*
 * execute walk if word is valid [speed]walk sequence -
 * return 1 if walked, 0 if not
 */
int parse_walk(word)
char *word;
{
    static char wbuf[2] = { ' ', '\n' };
    int n = strlen(word);
  
    if(!n || !strchr("nsewud", word[n - 1]) ||
	    strspn(word, "nsewud0123456789") != n)
	return 0;

    for( ; *word; word++) {
	n = isdigit(*word) ? strtol(word, &word, 10) : 1;
	for(wbuf[0] = *word; n--; ) {
	    send_to_host(wbuf, 2);
	    if(is_main_sess())
		add_dir_to_map(*word);
	}
    }
    return 1;
}

/*
 * verify that the braces in the line are properly balanced
 */
int brace_balance(str)
char *str;
{
    char *message = "#unbalanced %c near \"%.16s\"\n";
    char *d, *s = str;
    if(skip_arg(s, '\0')) return 1;
    if((d = skip_arg(s, '}')))
	printf(message, *d, (d - str > 15) ? d - 15 : str);
    else {
	do {
	    d = skip_arg(s, '{');
	    if(!d) return 1;
	    s = skip_arg(d + 1, '}');
	} while(s++);
	printf(message, *d, d);
    }
    return 0;
}

#if 0
/*
 * verify the label and command
 * for use with #action #alias #bind #identify #init #record #schedule...
 */
int valid_label_cmd(msg, lbl, cmd)
char *msg, *lbl, *cmd;
{
    if(lbl && skip_arg(lbl, ' '))
	printf("#%s label should not contain a space.\n", msg);
    else if(cmd && !skip_arg(cmd, '\0'))
	printf("#%s: unbalanced braces in command.\n", msg);
    else
	return 1;
    return 0;
}
#endif

/*
 * {} *must* be balanced in input string. input will be thrashed.
 * leave {...} contents intact, otherwise unescape and replace
 * &0 .. &9, &label, &{label}, &[expression], &(command).
 */
static void subst_param(buf, str, param)
char *buf, *str;
struct param param[10];
{
    char *src, *s = str;
    int len, room = LINELEN;

    while(*s) {
	if(*s != '&') {
	    if(*s != '{') {
		if(*s == '\\' && !*++s) break;
		if(room) --room, *buf++ = *s++;
		continue;
	    } else {
		src = s;  s = skip_arg(s + 1, '}') + 1;
		len = s - src;
	    }
	} else switch(*++s) {
	case '&':				/* delayed &	*/
	    src = s;  for(len = 1; *++s == '&'; len++);
	    break;
	case '{':				/* &{variable}	*/
	    s = skip_arg((src = s + 1), '}');
#ifndef XEXPR
	    {		/* {} must be balanced, no check here	*/
		char sbuf[LINELEN];
		*s++ = 0;  subst_param(sbuf, src, param);
		src = getenv(sbuf);
	    }
#else
	    src = brace_var(src);
#endif
	    if(!src) continue;
	    len = strlen(src);
	    break;
	case '[':				/* &[expr]	*/
	    s = skip_arg((src = s + 1), ']');
	    if(*s) {
		static char vbuf[24];  long val;
#ifndef XEXPR
		char sbuf[LINELEN];
		*s++ = 0;  subst_param(sbuf, src, param);
		if(!(src = calc(sbuf, &val)) || *src) continue;
#else
		*s++ = 0;
		if(!(src = calc(src, &val)) || *src) continue;
#endif
		len = sprintf(src = vbuf, "%ld", val);
		break;
	    } else
		printf("#unbalanced [ near \"&[%.16s\"\n", src);
	    continue;
	case '(':				/* &(command)	*/
	    s = skip_arg((src = s + 1), ')');
	    if(*s) {
		FILE *fp;
		*s++ = 0;
		unescape(src);
		if((fp = popen(src, "r"))) {
		    if((len = fread(buf, 1, room, fp)) > 0) {
			if(buf[len - 1] == '\n') len--;
			buf += len;  room -= len;
		    }
		    pclose(fp);
		}
	    } else
		printf("#unbalanced ( near \"&(%.16s\"\n", src);
	    continue;
	default:				/* &N or &label	*/
	    if(isdigit(*s)) {
		len = ctoi(*s++);
		if(!param || !param[len].len) continue;
		src = param[len].s;  len = param[len].len;
	    } else {
		for(src = s; islabel(*s); s++);
		len = *s;  *s = 0;
		src = getenv(src);  *s = len;
		if(!src) continue;
		len = strlen(src);
	    }
	} /* switch ends */
	if(room < len) len = room;
	memcpy(buf, src, len);
	buf += len;  room -= len;
    } /* end while */
    if(room) *buf = 0;
}

#ifdef XEXPR

#define T_NOP	0
#define T_MUL	1
#define	T_DIV	2
#define T_MOD	3
#define T_ADD	4
#define T_SUB	5
#define T_LT	6
#define T_GT	7
#define T_AND	8
#define T_OR	9
#define T_XOR	10
#define T_EQ	11
#define T_NE	12
#define T_SHL	13
#define T_SHR	14
#define T_LE	15
#define T_GE	16
#define T_LAND	17
#define T_LOR	18
#define N_PR	11		/* number of precedence levels */
static char precedence[] = { 0,10,10,10,9,9,7,7,5,3,4,6,6,8,8,7,7,2,1 };
static char *opstring = "*/%+-<>&|^=!";

static char *get_tok(str, op)
char *str, *op;
{
    char *s;
    int c = 0;
    if(*str && (s = strchr(opstring, *str))) {
	c = s - opstring + 1;  str++;
	switch(c) {
	    case T_EQ:
	    case T_NE:
		if(*str == '=') str++;
		else c = T_NOP;
		break;
	    case T_LT:
		if(*str == '<') { c = T_SHL;  str++; }
		else if(*str == '=') { c += T_LE - T_LT;  str++; }
		break;
	    case T_GT:
		if(*str == '>') { c = T_SHR;  str++; }
		else if(*str == '=') { c += T_GE - T_GT;  str++; }
		break;
	    case T_AND:
		if(*str == '&') { c += T_LAND - T_AND;  str++; }
		break;
	    case T_OR:
		if(*str == '|') { c += T_LOR - T_OR;  str++; }
		break;
	}
    }
    *op = c;
    return str;
}

static char *brace_var(str)
char *str;
{
    char buf[LINELEN];
    buf[sizeof(buf) - 1] = 0;
    subst_param(buf, str, (struct param *)0);
    return buf[sizeof(buf) - 1] ? 0 : getenv(buf);
}
#endif

/*
 * read a value from string; recurse into calc() if parentheses are found.
 * return string pointer properly advanced, NULL signifies failure.
 */
static char *getval(str, val)
char *str; long *val;
{
#ifndef XEXPR
    char *s = str;
    if(*s == '(') {
	s = calc(s + 1, val);
	return (s && *s++ == ')') ? s : 0;
    }
    *val = strtol(s, &str, 10);
    return s != str ? str : 0;
#else
    char *s = str, *k = 0;
    if(*s == '(') {
	s = calc(s + 1, val);
	return (s && *s++ == ')') ? s : 0;
    } else if(!isdigit(*s)) {
	if(*s == '&') s++;
        if(*s == '{') {
	    k = skip_arg(++s, '}');  *k++ = 0;
	    s = brace_var(s);
	    if(!s) return 0;
	} else {
	    char c;
	    k = s;  while(islabel(*k)) k++;
	    if(k == s) return 0;
	    c = *k; *k = 0; s = getenv(s); *k = c;
	}
    }
    *val = strtol(s, &str, 10);
    return s == str ? 0 : k ? k : str;
#endif
}

/*
 * calculate the value of a simple expression. () + - * / % are recognized.
 * return advanced string pointer, or NULL in case of syntax error.
 */
static char *calc(str, val)
char *str; long *val;
{
#ifndef XEXPR
    long v1, v2;
    char op, sign = 0;
    char *s = str;
    if(*s == '-') sign = 1, s++;
    for(*val = 0; (s = getval(s, &v1)); s++) {
	while(op = *s, op == '*' || op == '/' || op == '%')
	    if(!(s = getval(s + 1, &v2))) return 0;
	    else if(op == '*') v1 *= v2;
	    else if(!v2) return 0;
	    else if(op == '/') v1 /= v2;
	    else v1 %= v2;
	*val += sign ? -v1 : v1;
	if(op == '-') sign = 1;
	else if(sign = 0, op != '+') return s;
    }
    return 0;
#else /* XEXPR */
    long v[N_PR], vn1;
    char op[N_PR];
    char *u_s, *s = str;
    int u_ops, o1, o2, n = 0;
    do {
	for(u_s = s; *s == '+' || *s == '-' || *s == '~' || *s == '!'; s++);
	u_ops = s - u_s;
	s = getval(s, v + n);
	if(!s) break;
	/* HANDLE UNARY OPERATORS */
	while(u_ops)
	    if(u_s[--u_ops] == '-') v[n] = -v[n];
	    else if(u_s[u_ops] == '~') v[n] = ~v[n];
	    else if(u_s[u_ops] == '!') v[n] = !v[n];
	s = get_tok(s, op + n);
	/* HANDLE BINARY OPERATORS */
	while(n && precedence[o2 = op[n]] <= precedence[o1 = op[n - 1]])
	    switch(op[--n] = o2, vn1 = v[n + 1], o1) {
		case T_EQ: v[n] = v[n] == vn1;  break;
		case T_NE: v[n] = v[n] != vn1;  break;
		case T_LE: v[n] = v[n] <= vn1;  break;
		case T_GE: v[n] = v[n] >= vn1;  break;
		case T_LT: v[n] = v[n] < vn1;  break;
		case T_GT: v[n] = v[n] > vn1;  break;
		case T_LOR: v[n] = v[n] || vn1;  break;
		case T_LAND: v[n] = v[n] && vn1;  break;
		case T_SHL: v[n] <<= vn1;  break;
		case T_SHR: v[n] >>= vn1;  break;
		case T_OR: v[n] |= vn1;  break;
		case T_XOR: v[n] ^= vn1;  break;
		case T_AND: v[n] &= vn1;  break;
		case T_ADD: v[n] += vn1;  break;
		case T_SUB: v[n] -= vn1;  break;
		case T_MUL: v[n] *= vn1;  break;
		case T_DIV: v[n] /= vn1;  break;
		case T_MOD: v[n] %= vn1;  break;
	    }
    } while(op[n++]);
    *val = v[0];
    return s;
#endif
}

/*
 * match action containing &0..&9 and return actual params in param array
 *
 * FEATURE: & should not match an empty word.
 * FEATURE: in pattern in a sequence of &n-s and/or &-s, all but last are
 * ignored. such pattern makes little sense anyway.
 * FEATURE: no backtracking used, therefore a match with & might fail, even
 * though respective regexp would not, e.g. pat:"a & b", src:"a a a b" fails.
 */
static struct param *match_action(pat, src)
char *pat, *src;
{
    static struct param param[10], *pp;
    static int dirtyregs;
    char mpat[LINELEN], *s;
    int mbeg = 0, mword, mregs, prm, k;
    if(*pat == '^') {		/* anchor match at line start */
	if(col0) return 0;
	pat++;  mbeg = 1;
    }
    if(*pat == '&')		/* pattern starts with '&' */
	mbeg--;
    mword = mregs = 0;  prm = -1;  param[0].s = src;
    while(*pat) {
	if(*pat == '&') do {	/* pick out parameters */
	    pat++;  prm = ctoi(*pat++);
	    if((unsigned)prm >= 10)	/* support single & for back compat */
		pat--, prm = 0, mword = 1;	/* match a single word only */
	} while(*pat == '&');
	s = src;
	if(*pat) {		/* attempt match */
	    char *mp = mpat;
	    do *mp++ = *pat++; while(*pat && *pat != '&');
	    if(mbeg > 0 ? strncmp(src, mpat, mp - mpat) :
		    (*mp = '\0', src = strstr(src, mpat), !src)) {
		dirtyregs |= mregs;
		return 0;
	    }
	    if(prm != -1) {
		mregs |= 1 << prm;	/* mark the parameter as set */
		param[prm].len = src - (param[prm].s = s);
	    }
	    src += mp - mpat;
	    k = mbeg;  mbeg = 0; 
	    if(!mword) continue;

	    /* post-process single word match */
	    mword = 0;
	    if(k != -1) {	/* match only if param is single-worded */
		for(k = param[0].len; k; ) if(k--, !isalpha(*s++)) {
		    dirtyregs |= mregs;
		    return 0;
		}
	    } else {		/* unanchored '&' start, take last word */
		for(k = param[0].len; k && isalpha(s[k - 1]); k--);
		param[0].len -= k;  param[0].s += k;
	    }
	    continue;
	}
	if(prm != -1) {		/* '&' at the end of the pattern space */
	    if(mword) {		/* take first word */
		if(mbeg == -1) while(*s && !isalpha(*s)) s++;
		for(k = 0; isalpha(s[k]); k++);
	    } else
		k = strlen(s);
	    mregs |= 1 << prm;
	    param[prm].len = k;  param[prm].s = s;
	}
	break;
    } /* end while */
    if(!(mregs & 1))
	param[0].len = src + strlen(src) - param[0].s;
    k = dirtyregs & ~mregs;  dirtyregs = mregs;  pp = param;
    while((k >>= 1))
	if(pp++, k & 1) pp->len = 0;
    return param;
}

/*
 * Search for actions to trig on an input line. The line can't be trashed
 * since we want to print it on the screen later.
 * Return 1 if line should be printed, 0 if it should be suppressed.
 */
int search_action(line)
char *line;
{
    actionnode *p;
    for(p = actions; p; p = p->next)
	if(p->active) {
	    struct param *match = match_action(p->pattern, line);
	    if(!match) continue;
	    if(strcmp("#gag", p->commands) == 0) return 0;
	    clear_prompt();
	    parse(p->commands, 0, match);
	    redraw_prompt();
	    break;
	}
    return 1;
}

/*
 * When needed, enclose in {} and escape &\; with a backslash.
 */
void escape_command(str, buf)
char str[LINELEN], buf[LINELEN];
{
    char *s;
    if(skip_arg(str, '\0') && !skip_arg(str, CMDSEP)) {
	strcpy(buf, str);
    } else {
	char *k, *d = buf;
	s = k = str;  *d++ = '{';
	do {
	    if(s == k) {
		if(*s == CMDSEP || *s == '&' ||
			(*s != '\\' ?
			    (*s == '{' && !(k = skip_arg(s + 1, '}'))
			     && (k = s, 1))
			:
			    (!*(s + 1) || (*d++ = '\\', k++, 1)))) {
		    if(d - buf >= LINELEN - 2) break;
		    *d++ = '\\';
		}
		k++;
	    }
	    *d++ = *s++;
	} while(d - buf < LINELEN && (*s || (*d++ = '}', 0)));
	*d = '\0';
    }
}

/*
 * escape special chars with a backslash
 */
void escape_specials(str, buf, spec)
char *str, *buf, *spec;
{
    int k;
    char *s, *d;
    for(s = str, d = buf; ; *d++ = '\\', *d++ = *s++) {
	if((k = strcspn(s, spec)))
	    do *d++ = *s++; while(--k);
	if(!*s) break;
    }
    *d = 0;
}

/*
 * remove escapes (backslashes) from a string
 */
void unescape(s)
char *s;
{
    char *p;
    for(p = s; *s && (*s != '\\' || *++s); *p++ = *s++);
    *p = 0;
}

/*
 * split str into arguments separated by whitespace.
 * parameter 0 contains the whole line.
 */
static void split_arg(str, param)
char *str;
struct param param[10];
{
    int i = 1;
    char *s = str;
    param[0].s = s;
    param[0].len = strlen(s);
    do {
	param[i].s = s = skipspace(s);
	if(!*s) break;
	while(*++s && *s != ' ');
	param[i].len = s - param[i].s;
    } while(++i < 10 && *s++);
    while(i < 10) param[i++].len = 0;
}

/*
 * like strchr, but honour (possibly nested) braces and escaping backslashes.
 * return pointer to the delim in str, or NULL.
 */
static char *skip_arg(str, delim)
char *str;
char delim;
{
    char *s = str - 1;
    for( ; ; ) {
	if(*++s == delim) return s;
	if(!*s || *s == '}') break;
	if(*s != '\\') {
	    if(*s == '{') {
		int nested = 1;
		for( ; ; )	/* inner loop (text within {}) */
		    if(*++s)
			if(*s != '{') {
			    if(*s != '}') {
				if(*s == '\\' && !*++s) return 0;
			    } else if(!--nested) break;
			} else nested++;
		    else return 0;
	    }
	} else if(!*++s) break;
    }
    return 0;
}

/*
 * split the argument into left and right sides, separated by delim
 * return pointer to right hand side, or 0 if no delim was found.
 */
char *chop_arg(str, delim)
char *str, delim;
{
    char *s = skip_arg(str, delim);
    if(s) *s++ = 0;
    return s;
}

/*
 * like chop_arg, but no special handling of {\;
 * used as a replacement of strtok
 */
char *chop_str(str, delim)
char *str, delim;
{
    char *s = str;
    if(*s) while(*s != delim && *++s);
    if(*s) for(*s = 0; *++s == delim; );
    return *s ? s : 0;
}

/*
 * return pointer to next non-blank char
 * FEATURE: only space is considered whitespace.
 * editline is all printable chars anyway.
 */
char *skipspace(p)
char *p;
{
    while(*p == ' ') p++;
    return p;
}

/*
 * return pointer to next non-digit
 */
char *skipnum(p)
char *p;
{
    if(*p == '-') p++;
    while(isdigit(*p)) p++;
    return p;
}

